/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

/*
begin_html
See also:
<ul>
<li><a href="ublas_matrixutil.h.html">ublas_matrixutil.h</a></li>  
<li><a href="matrixutil.h.html">matrixutil.h</a></li>
</ul>
<b>Up:</b> <a href="index.html">Index</a>
end_html
*/

#ifdef _WIN32
  #include <windows.h>
#else
  #include <stdlib.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <string.h>
#include <iostream>
#include <algorithm>

#include "ublas_plus.h"
#include "ublas_matrixutil.h"

using std::cerr;
using std::endl;
using std::min;

void invm(matrix<double>& Q_in, matrix<double>& Q_out)
{
  assert( Q_in.size1() == Q_in.size2() );
  int n = Q_in.size1();
  fq::matrix<double> Q(n, n);
  matrix_to_arrays(Q_in, (double**)Q, n, n);
  gaussj0(Q, n, NULL, 0, cerr);
  if ( Q_out.size1() != n || Q_out.size2() != n )
    Q_out.resize(n, n);
  matrix_from_arrays(Q_out, (double**)Q, n, n);
}
  
bool eigen_vals_vecs(matrix<double>& Q_in, double *rvalues, double *ivalues, matrix<double>& vectors_out, matrix<double>& vectors_inv, std::ostream& output)
{
  assert( Q_in.size1() == Q_in.size2() );
  int n = Q_in.size1();
  fq::matrix<double> Q(n, n);
  matrix_to_arrays(Q_in, (double**)Q, n, n);
  fq::vector<double> r_1base(n+1), i_1base(n+1);
  fq::matrix<double> V(n, n);
  bool result = eigen((double**)Q, n, (double*)r_1base, (double*)i_1base, V, output);
  if ( vectors_out.size1() != n || vectors_out.size2() != n )
    vectors_out.resize(n, n);
  if ( vectors_inv.size1() != n || vectors_inv.size2() != n )
    vectors_inv.resize(n, n);
  matrix_from_arrays(vectors_out, (double**)V, n, n);
  invm(vectors_out, vectors_inv);
  memcpy(rvalues, r_1base+1, n*sizeof(double));
  memcpy(ivalues, i_1base+1, n*sizeof(double));
  return result;
}

void expm_eig(double *rvalues, matrix<double>& vectors, matrix<double>& vectors_inv, double t,
	      matrix<double>& tmp_nxn_diag, matrix<double>& tmp_nxn, matrix<double>& Q_out)
{
  int n = vectors.size1();
  for (int i=0; i<n; ++i) {
    tmp_nxn_diag(i,i) = exp(rvalues[i] * t);
  }
  
  noalias(tmp_nxn) = prod(vectors, tmp_nxn_diag);
  noalias(Q_out) = prod(tmp_nxn, vectors_inv);
}  

void expm(matrix<double>& Q_in, double t, matrix<double>& Q_out)
{
  int n = Q_in.size1();
  fq::vector<double> rvalues(n), ivalues(n);
  matrix<double> vectors;
  matrix<double> vectors_inv;
  matrix<double> tmp_nn(n, n);
  matrix<double> tmp_diag(n, n);
  for (int i=0; i<n; ++i)
    for (int j=0; j<n; ++j)
      tmp_diag(i,j) = 0.0;
  

  eigen_vals_vecs(Q_in, rvalues, ivalues, vectors, vectors_inv, cerr);
  expm_eig(rvalues, vectors, vectors_inv, t, tmp_diag, tmp_nn, Q_out);
}

#define TINY 1e-8
  
/* SVD like linpack:
#   * V is transposed
# but all matrices have full dimension
# Then we repair and check it:
#   * tiny numbers are replaced with 0
#   * U and V need orthonormal columns
*/

bool sv_decomp_inner(double** uu, int Nr, int Nc, matrix<double>& U, matrix<double>& W, matrix<double>& Vi);

bool sv_decomp(matrix<double>& A, matrix<double>& U, matrix<double>& W, matrix<double>& Vi)
{
  int Nr = A.size1();
  int Nc = A.size2();
  fq::matrix<double> uu(Nr+1, Nc+1);
  matrix_to_arrays(A, (double**) uu, Nr, Nc);
  return sv_decomp_inner(uu, Nr, Nc, U, W, Vi);
}

bool sv_decomp(double** A, int Nr, int Nc, matrix<double>& U, matrix<double>& W, matrix<double>& Vi)
{
  fq::matrix<double> uu(Nr+1, Nc+1);
  for (int i=0; i<Nr; ++i)
    for (int j=0; j<Nc; ++j)
      uu[i][j] = A[i][j];
  return sv_decomp_inner(uu, Nr, Nc, U, W, Vi);
}

bool sv_decomp_inner(double** uu, int Nr, int Nc, matrix<double>& U, matrix<double>& W, matrix<double>& Vi)
{  
  // automatic storage
  fq::vector<double> _ww(Nc+1);
  fq::matrix<double> _vv(Nc+1, Nc+1);
  // as pointers
  double *ww = _ww, **vv = _vv;
  // SVD
  svdecomp(uu, Nr, Nc, ww, vv, cerr);
  // outputs
  if ( U.size1() != Nr || U.size2() != Nc )
    U.resize(Nr, Nc);
  for (int i=0; i<Nr; ++i)
    for (int j=0; j<Nc; ++j)
      U(i, j) = (fabs(uu[i][j]) > TINY) ? uu[i][j] : 0.0;
  if ( W.size1() != Nc || W.size2() != Nc )
    W.resize(Nc, Nc);
  for (int i=0; i<Nc; ++i)
    for (int j=0; j<Nc; ++j)
      W(i, j) = ((i == j) && (fabs(ww[i]) > TINY)) ? ww[i] : 0.0;
  if ( Vi.size1() != Nc || Vi.size2() != Nc )
    Vi.resize(Nc, Nc);
  for (int i=0; i<Nc; ++i)
    for (int j=0; j<Nc; ++j)
      Vi(i, j) = (fabs(vv[i][j]) > TINY) ? vv[i][j] : 0.0;

  // checks a la MIL
  double s;
  for (int i=0; i<Nr; i++) {
    for (int j=Nr; j<Nc; j++) {
      if(fabs(uu[i][j])>TINY) {
        return false;
      }
    }
  }
  int Nlo = min(Nr, Nc);
  for (int i=0; i<Nlo; i++) { // make sure that u's columns are orthogonal unit vectors
    for (int j=0; j<Nlo; j++) { // changed from nx to mc Chris 6/2/2
      s=0.0;
      for (int k=0; k<Nr; k++) 
        s=s+uu[k][i]*uu[k][j];
      if( (i!=j&&fabs(s)>TINY) || (i==j&&fabs(s-1.0)>TINY) ) {
        return false;
      }
    }
  }
  for (int i=0; i<Nc; i++) { // make sure that v's columns are orthogonal unit vectors
    for (int j=0; j<Nc; j++) {
      s=0.0;
      for (int k=0; k<Nc; k++) 
        s=s+vv[k][i]*vv[k][j];
      if( (i!=j && fabs(s)>TINY) || (i==j && fabs(s-1.0)>TINY) ){
        return false;
      }
    }
  }
  for (int i=0; i<Nc; i++) {
    for (int j=0; j<Nc; j++) {
      s=0.0;
      for (int k=0; k<Nc; k++) 
        s=s+vv[i][k]*vv[j][k];
      if( (i!=j&&fabs(s)>TINY) || (i==j&&fabs(s-1.0)>TINY) ){
        return false;
      }
    }
  }  
  
  return true;
}

bool sv_invm(double** A, int n, double** Ai)
{
	matrix<double> U(n, n), W(n, n), Vi(n, n);
	bool result = sv_decomp(A, n, n, U, W, Vi);
	for (int i=0; i<n; ++i) {
		if ( W(i,i) )
			W(i,i) = 1.0 / W(i,i);
	}
	matrix<double> out(n, n);
	noalias(out) = prod(Vi, matrix<double>(prod(W, trans(U))));
	matrix_to_arrays(out, Ai, n, n);
	return result;
}
